from __future__ import annotations

import logging
from collections import UserDict
from typing import Any

from coredis.typing import (
    Hashable,
    Iterable,
    Mapping,
    MutableMapping,
    ResponseType,
    StringT,
    TypeVar,
)

logger = logging.getLogger("coredis")
logger.addHandler(logging.NullHandler())

T = TypeVar("T")
U = TypeVar("U")


class EncodingInsensitiveDict(UserDict[Any, Any]):
    def __init__(
        self,
        initial: MutableMapping[Any, Any] | None = None,
        encoding: str = "utf-8",
    ):
        self._encoding = encoding
        super().__init__(initial or {})

    def _alt_key(self, key: StringT) -> StringT:
        if isinstance(key, str):
            try:
                byte_alt = key.encode(self._encoding)
                if byte_alt in self.data:
                    return byte_alt
            except UnicodeEncodeError:
                pass
        elif isinstance(key, bytes):
            try:
                str_alt = key.decode(self._encoding)
                if str_alt in self.data:
                    return str_alt
            except UnicodeDecodeError:
                pass
        return key

    def __getitem__(self, key: StringT) -> Any:
        if key in self.data:
            return self.data[key]
        alt = self._alt_key(key)
        if alt in self.data:
            return self.data[alt]
        raise KeyError(key)

    def __setitem__(self, key: StringT, value: Any) -> None:
        alt = self._alt_key(key)
        self.data[alt] = value

    def __delitem__(self, key: StringT) -> None:
        alt = self._alt_key(key)
        del self.data[alt]

    def __contains__(self, key: Any) -> bool:
        return key in self.data or self._alt_key(key) in self.data

    def get(self, key: StringT, default: Any = None) -> Any:
        return self.data.get(key, self.data.get(self._alt_key(key), default))

    def pop(self, key: StringT, default: Any = None) -> Any:
        if key in self.data:
            return self.data.pop(key)
        alt = self._alt_key(key)
        return self.data.pop(alt, default)

    def stringify_keys(self) -> dict[str, Any]:
        d = {}
        for key, value in self.items():
            d[key.decode(self._encoding) if isinstance(key, bytes) else key] = (
                value.stringify_keys() if isinstance(value, EncodingInsensitiveDict) else value
            )
        return d


def b(x: ResponseType, encoding: str | None = None) -> bytes:
    if isinstance(x, bytes):
        return x
    if not isinstance(x, str):
        _v = str(x)
    else:
        _v = x
    return _v.encode(encoding) if encoding else _v.encode()


def nativestr(x: ResponseType, encoding: str = "utf-8") -> str:
    if isinstance(x, (str, bytes)):
        return x if isinstance(x, str) else x.decode(encoding, "replace")
    elif isinstance(x, (int, float, bool)):
        return str(x)
    raise ValueError(f"Unable to cast {x} to string")


def tuples_to_flat_list(nested_list: Iterable[tuple[T, ...]]) -> list[T]:
    return [item for sublist in nested_list for item in sublist]


def dict_to_flat_list(mapping: Mapping[T, U], reverse: bool = False) -> list[T | U]:
    e1: list[T | U] = list(mapping.keys())
    e2: list[T | U] = list(mapping.values())

    ret: list[T | U] = []

    if reverse:
        e1, e2 = e2, e1

    for idx, e in enumerate(e1):
        ret.append(e)
        ret.append(e2[idx])

    return ret


def make_hashable(*args: Any) -> tuple[Hashable, ...]:
    return tuple(
        (
            tuple(make_hashable(v)[0] for v in a)
            if isinstance(a, (tuple, list))
            else (
                frozenset(make_hashable(v)[0] for v in a)
                if isinstance(a, set)
                else (
                    tuple((k, make_hashable(v)[0]) for k, v in a.items())
                    if isinstance(a, dict)
                    # this will fail downstream if `a` is not hashable
                    else a
                )
            )
        )
        for a in args
    )


def query_param_to_bool(value: Any) -> bool | None:
    if value is None or value in ("", b""):
        return None
    if isinstance(value, (int, float, bool, str, bytes)):
        value = nativestr(value)
        if value.upper() in ("0", "F", "FALSE", "N", "NO"):
            return False
        elif value.upper() in ("1", "T", "TRUE", "Y", "YES"):
            return True

    return bool(value)


# ++++++++++ cluster utils ++++++++++++++

x_mode_m_crc16_lookup = [
    0x0000,
    0x1021,
    0x2042,
    0x3063,
    0x4084,
    0x50A5,
    0x60C6,
    0x70E7,
    0x8108,
    0x9129,
    0xA14A,
    0xB16B,
    0xC18C,
    0xD1AD,
    0xE1CE,
    0xF1EF,
    0x1231,
    0x0210,
    0x3273,
    0x2252,
    0x52B5,
    0x4294,
    0x72F7,
    0x62D6,
    0x9339,
    0x8318,
    0xB37B,
    0xA35A,
    0xD3BD,
    0xC39C,
    0xF3FF,
    0xE3DE,
    0x2462,
    0x3443,
    0x0420,
    0x1401,
    0x64E6,
    0x74C7,
    0x44A4,
    0x5485,
    0xA56A,
    0xB54B,
    0x8528,
    0x9509,
    0xE5EE,
    0xF5CF,
    0xC5AC,
    0xD58D,
    0x3653,
    0x2672,
    0x1611,
    0x0630,
    0x76D7,
    0x66F6,
    0x5695,
    0x46B4,
    0xB75B,
    0xA77A,
    0x9719,
    0x8738,
    0xF7DF,
    0xE7FE,
    0xD79D,
    0xC7BC,
    0x48C4,
    0x58E5,
    0x6886,
    0x78A7,
    0x0840,
    0x1861,
    0x2802,
    0x3823,
    0xC9CC,
    0xD9ED,
    0xE98E,
    0xF9AF,
    0x8948,
    0x9969,
    0xA90A,
    0xB92B,
    0x5AF5,
    0x4AD4,
    0x7AB7,
    0x6A96,
    0x1A71,
    0x0A50,
    0x3A33,
    0x2A12,
    0xDBFD,
    0xCBDC,
    0xFBBF,
    0xEB9E,
    0x9B79,
    0x8B58,
    0xBB3B,
    0xAB1A,
    0x6CA6,
    0x7C87,
    0x4CE4,
    0x5CC5,
    0x2C22,
    0x3C03,
    0x0C60,
    0x1C41,
    0xEDAE,
    0xFD8F,
    0xCDEC,
    0xDDCD,
    0xAD2A,
    0xBD0B,
    0x8D68,
    0x9D49,
    0x7E97,
    0x6EB6,
    0x5ED5,
    0x4EF4,
    0x3E13,
    0x2E32,
    0x1E51,
    0x0E70,
    0xFF9F,
    0xEFBE,
    0xDFDD,
    0xCFFC,
    0xBF1B,
    0xAF3A,
    0x9F59,
    0x8F78,
    0x9188,
    0x81A9,
    0xB1CA,
    0xA1EB,
    0xD10C,
    0xC12D,
    0xF14E,
    0xE16F,
    0x1080,
    0x00A1,
    0x30C2,
    0x20E3,
    0x5004,
    0x4025,
    0x7046,
    0x6067,
    0x83B9,
    0x9398,
    0xA3FB,
    0xB3DA,
    0xC33D,
    0xD31C,
    0xE37F,
    0xF35E,
    0x02B1,
    0x1290,
    0x22F3,
    0x32D2,
    0x4235,
    0x5214,
    0x6277,
    0x7256,
    0xB5EA,
    0xA5CB,
    0x95A8,
    0x8589,
    0xF56E,
    0xE54F,
    0xD52C,
    0xC50D,
    0x34E2,
    0x24C3,
    0x14A0,
    0x0481,
    0x7466,
    0x6447,
    0x5424,
    0x4405,
    0xA7DB,
    0xB7FA,
    0x8799,
    0x97B8,
    0xE75F,
    0xF77E,
    0xC71D,
    0xD73C,
    0x26D3,
    0x36F2,
    0x0691,
    0x16B0,
    0x6657,
    0x7676,
    0x4615,
    0x5634,
    0xD94C,
    0xC96D,
    0xF90E,
    0xE92F,
    0x99C8,
    0x89E9,
    0xB98A,
    0xA9AB,
    0x5844,
    0x4865,
    0x7806,
    0x6827,
    0x18C0,
    0x08E1,
    0x3882,
    0x28A3,
    0xCB7D,
    0xDB5C,
    0xEB3F,
    0xFB1E,
    0x8BF9,
    0x9BD8,
    0xABBB,
    0xBB9A,
    0x4A75,
    0x5A54,
    0x6A37,
    0x7A16,
    0x0AF1,
    0x1AD0,
    0x2AB3,
    0x3A92,
    0xFD2E,
    0xED0F,
    0xDD6C,
    0xCD4D,
    0xBDAA,
    0xAD8B,
    0x9DE8,
    0x8DC9,
    0x7C26,
    0x6C07,
    0x5C64,
    0x4C45,
    0x3CA2,
    0x2C83,
    0x1CE0,
    0x0CC1,
    0xEF1F,
    0xFF3E,
    0xCF5D,
    0xDF7C,
    0xAF9B,
    0xBFBA,
    0x8FD9,
    0x9FF8,
    0x6E17,
    0x7E36,
    0x4E55,
    0x5E74,
    0x2E93,
    0x3EB2,
    0x0ED1,
    0x1EF0,
]


def crc16(data: bytes) -> int:
    crc = 0

    for byte in data:
        crc = ((crc << 8) & 0xFF00) ^ x_mode_m_crc16_lookup[((crc >> 8) & 0xFF) ^ byte]

    return crc & 0xFFFF


def hash_slot(key: bytes) -> int:
    start = key.find(b"{")

    if start > -1:
        end = key.find(b"}", start + 1)

        if end > -1 and end != start + 1:
            key = key[start + 1 : end]

    return crc16(key) % 16384


__all__ = [
    "hash_slot",
    "logger",
    "EncodingInsensitiveDict",
]



from __future__ import annotations

import dataclasses
import time
import weakref
from abc import ABC, abstractmethod
from collections import Counter
from contextlib import AsyncExitStack
from typing import TYPE_CHECKING, Any, cast

from anyio import (
    TASK_STATUS_IGNORED,
    create_task_group,
    sleep,
)
from anyio.abc import TaskStatus

from coredis._utils import b, make_hashable
from coredis.commands.constants import CommandName
from coredis.connection._base import BaseConnection, ConnectionT
from coredis.connection._cluster import ClusterConnection
from coredis.connection._tcp import TCPLocation
from coredis.exceptions import ConnectionError
from coredis.retry import ExponentialBackoffRetryPolicy
from coredis.typing import (
    Generic,
    OrderedDict,
    RedisValueT,
    ResponseType,
    StringT,
)

if TYPE_CHECKING:
    import coredis.client
    from coredis.pool._base import BaseConnectionPool


@dataclasses.dataclass
class CacheStats:
    """
    Summary of statics to be used by instances of :class:`~coredis.patterns.cache.AbstractCache`
    The individual counters exposed are not guaranteed to retain fine grained per key
    metrics but the totals (returned by :attr:`coredis.patterns.cache.CacheStats.summary`) will be maintained
    aggregated.
    """

    #: summary of hits by key (for all commands)
    hits: Counter[bytes] = dataclasses.field(default_factory=Counter)
    #: summary of misses by key (for all commands)
    misses: Counter[bytes] = dataclasses.field(default_factory=Counter)
    #: number of invalidations including server side and local invalidations
    invalidations: Counter[bytes] = dataclasses.field(default_factory=Counter)
    #: counter of keys which returned dirty results based on confidence testing
    dirty: Counter[bytes] = dataclasses.field(default_factory=Counter)

    def clear(self) -> None:
        self.hits.clear()
        self.misses.clear()
        self.invalidations.clear()
        self.dirty.clear()

    def compact(self) -> None:
        """
        Collapse totals into a single key to avoid unbounded growth of stats

        :meta private:
        """

        for counter in [self.hits, self.misses, self.invalidations, self.dirty]:
            total = sum(counter.values())
            counter.clear()
            counter[b"__coredis__internal__stats__total"] = total

    def hit(self, key: RedisValueT) -> None:
        self.hits[b(key)] += 1

    def miss(self, key: RedisValueT) -> None:
        self.misses[b(key)] += 1

    def invalidate(self, key: RedisValueT) -> None:
        self.invalidations[b(key)] += 1

    def mark_dirty(self, key: RedisValueT) -> None:
        self.dirty[b(key)] += 1

    @property
    def summary(self) -> dict[str, int]:
        """
        Aggregated totals of ``hits``, ``misses``, ``dirty_hits``
        and ``invalidations``
        """

        return {
            "hits": sum(self.hits.values()),
            "misses": sum(self.misses.values()),
            "dirty_hits": sum(self.dirty.values()),
            "invalidations": sum(self.invalidations.values()),
        }

    def __repr__(self) -> str:
        summary = self.summary

        return (
            f"CacheStats<hits={summary['hits']}, "
            f"misses={summary['misses']}, "
            f"dirty_hits={summary['dirty_hits']}, "
            f"invalidations={summary['invalidations']}>"
        )


class AbstractCache(ABC):
    """
    Abstract class representing a local cache that can be used by
    :class:`~coredis.Redis` or :class:`coredis.RedisCluster`
    """

    @abstractmethod
    def get(self, command: bytes, key: RedisValueT, *args: RedisValueT) -> ResponseType:
        """
        Fetch the cached response for command/key/args combination
        """
        ...

    @abstractmethod
    def put(
        self, command: bytes, key: RedisValueT, *args: RedisValueT, value: ResponseType
    ) -> None:
        """
        Cache the response for command/key/args combination
        """
        ...

    @abstractmethod
    def invalidate(self, *keys: RedisValueT) -> None:
        """
        Invalidate any cached entries for the provided keys
        """
        ...

    @abstractmethod
    def reset(self) -> None:
        """
        Reset the cache
        """
        ...

    @property
    @abstractmethod
    def stats(self) -> CacheStats:
        """
        Returns the current stats for the cache
        """
        ...

    @property
    @abstractmethod
    def confidence(self) -> float:
        """
        Confidence in cached values between 0 - 100. Lower values
        will result in the client discarding and / or validating the
        cached responses
        """
        ...

    @abstractmethod
    def feedback(self, command: bytes, key: RedisValueT, *args: RedisValueT, match: bool) -> None:
        """
        Provide feedback about a key as having either a match or drift from the actual
        server side value
        """
        ...

    @property
    @abstractmethod
    def healthy(self) -> bool:
        """
        Whether the cache is healthy and should be taken seriuosly
        """
        ...


class LRUCache(AbstractCache):
    def __init__(
        self,
        max_keys: int = 2**12,
        confidence: float = 100,
        dynamic_confidence: bool = False,
    ) -> None:
        """
        Implementation of an LRU cache that can be used
        with :class:`~coredis.Redis`  or :class:`coredis.RedisCluster`

        :param max_keys: maximum keys to cache. A negative value represents and unbounded cache.
        :param confidence: 0 - 100. Lower values will result in the client discarding
         and / or validating the cached responses
        :param dynamic_confidence: Whether to adjust the confidence based on sampled validations.
         Tainted values drop the confidence by 0.1% and confirmations of correct cached values
         will increase the confidence by 0.01% upto 100.
        """
        self._confidence = self._original_confidence = confidence
        self._dynamic_confidence = dynamic_confidence
        self._stats = CacheStats()
        self.max_keys = max_keys
        # key -> (command, args) -> response
        self._storage: OrderedDict[bytes, dict[tuple[bytes, Any], ResponseType]] = OrderedDict()

    def put(
        self, command: bytes, key: RedisValueT, *args: RedisValueT, value: ResponseType
    ) -> None:
        key_bytes = b(key)
        composite_key = (command, make_hashable(*args))

        if key_bytes not in self._storage and len(self._storage) >= self.max_keys:
            if self._storage:
                self._storage.popitem(last=False)

        # Get or create the key's cache dict
        if key_bytes not in self._storage:
            self._storage[key_bytes] = {}

        self._storage[key_bytes][composite_key] = value
        self._storage.move_to_end(key_bytes)

    def get(self, command: bytes, key: RedisValueT, *args: RedisValueT) -> ResponseType:
        key_bytes = b(key)
        if key_bytes not in self._storage:
            self._stats.miss(key)
            raise KeyError(key)

        # Move to end for LRU
        self._storage.move_to_end(key_bytes)
        composite_key = (command, make_hashable(*args))
        if composite_key not in self._storage[key_bytes]:
            self._stats.miss(key)
            raise KeyError(key)

        self._stats.hit(key)
        return self._storage[key_bytes][composite_key]

    def invalidate(self, *keys: RedisValueT) -> None:
        for key in keys:
            self._stats.invalidate(key)
            self._storage.pop(b(key), None)

    def reset(self) -> None:
        self._storage.clear()
        self._stats.compact()
        self._confidence = self._original_confidence

    @property
    def stats(self) -> CacheStats:
        return self._stats

    @property
    def confidence(self) -> float:
        return self._confidence

    def feedback(self, command: bytes, key: RedisValueT, *args: RedisValueT, match: bool) -> None:
        if not match:
            self._stats.mark_dirty(key)
            self.invalidate(key)

        if self._dynamic_confidence:
            self._confidence = min(
                100.0,
                max(0.0, self._confidence * (1.0001 if match else 0.999)),
            )

    @property
    def healthy(self) -> bool:
        """
        The LRU Cache is always "healthy"
        """
        return True


class TrackingCache(AbstractCache, Generic[ConnectionT]):
    """
    Abstract layout of a tracking cache to be used internally
    by coredis clients (Redis/RedisCluster)
    """

    _cache: AbstractCache

    def __init__(
        self, connection_pool: BaseConnectionPool[ConnectionT], cache: AbstractCache
    ) -> None:
        self._cache = cache
        self._retry_policy = ExponentialBackoffRetryPolicy(
            (ConnectionError,), retries=None, base_delay=1, max_delay=16, jitter=True
        )
        self._connection_pool = weakref.proxy(connection_pool)

    @abstractmethod
    async def run(self, task_status: TaskStatus[None] = TASK_STATUS_IGNORED) -> None:
        pass

    @abstractmethod
    def get_client_id(
        self,
        connection: ConnectionT,
    ) -> int | None:
        pass

    def get(self, command: bytes, key: RedisValueT, *args: RedisValueT) -> ResponseType:
        return self._cache.get(command, key, *args)

    def put(
        self, command: bytes, key: RedisValueT, *args: RedisValueT, value: ResponseType
    ) -> None:
        self._cache.put(command, key, *args, value=value)

    def invalidate(self, *keys: RedisValueT) -> None:
        self._cache.invalidate(*keys)

    def reset(self) -> None:
        self._cache.reset()

    @property
    def stats(self) -> CacheStats:
        return self._cache.stats

    @property
    def confidence(self) -> float:
        if not self.healthy:
            return 0
        return self._cache.confidence

    def feedback(self, command: bytes, key: RedisValueT, *args: RedisValueT, match: bool) -> None:
        self._cache.feedback(command, key, *args, match=match)


class NodeTrackingCache(TrackingCache[ConnectionT]):
    """
    Wraps an AbstractCache instance to use server assisted client caching
    to ensure local cache entries are invalidated if any operations are
    performed on the keys by another client.
    """

    _connection: ConnectionT | None

    def __init__(
        self,
        connection_pool: BaseConnectionPool[ConnectionT],
        cache: AbstractCache | None = None,
        max_idle_seconds: float = 15,
    ) -> None:
        """
        :param connection_pool: Connection pool used to acquire
         a connection for the tracking connection
        :param cache: AbstractCache instance to wrap
        :param max_idle_seconds: Maximum duration (in seconds) to tolerate no
         updates from the server before performing a keepalive check with a
         ``PING``. If no updates or successful ping occur within this period,
         the cache is considered unhealthy (as reflected by the ``healthy``
         property).
        """
        super().__init__(connection_pool, cache or LRUCache())
        self.client_id: int | None = None
        self._last_checkin: float = 0
        self._max_idle_seconds = max_idle_seconds
        self._connection = None

    def get_client_id(
        self,
        connection: BaseConnection,
    ) -> int | None:
        return self.client_id

    @property
    def healthy(self) -> bool:
        return bool(
            self._connection is not None
            and self._connection.usable
            and (time.monotonic() - self._last_checkin) < self._max_idle_seconds
        )

    async def run(self, task_status: TaskStatus[None] = TASK_STATUS_IGNORED) -> None:
        """
        Run a single connection that listens for invalidation messages,
        with reconnection logic.
        """
        started = False

        async def _run() -> None:
            nonlocal started
            async with self._connection_pool.acquire() as self._connection:
                if self._connection.tracking_client_id:
                    await self._connection.update_tracking_client(False)
                self.client_id = self._connection.client_id
                async with create_task_group() as self._tg:
                    self._tg.start_soon(self._consumer)
                    self._tg.start_soon(self._keepalive)
                    if not started:
                        task_status.started()
                        started = True
                        self._last_checkin = time.monotonic()
                    else:  # flush cache
                        self.reset()

        return await self._retry_policy.call_with_retries(_run)

    async def _keepalive(self) -> None:
        while True:
            if (idle := time.monotonic() - self._last_checkin) >= self._max_idle_seconds:
                if self._connection and await self._connection.create_request(CommandName.PING) in {
                    b"PONG",
                    "PONG",
                }:
                    self._last_checkin = time.monotonic()
            await sleep(max(1, self._max_idle_seconds - idle))

    async def _consumer(self) -> None:
        while True:
            if self._connection:
                async for response in self._connection.push_messages:
                    self._last_checkin = time.monotonic()
                    messages = cast(list[StringT], response[1] or [])
                    for key in messages:
                        self._cache.invalidate(key)


class ClusterTrackingCache(TrackingCache[ClusterConnection]):
    """
    An LRU cache for redis cluster that uses server assisted client caching
    to ensure local cache entries are invalidated if any operations are performed
    on the keys by another client.

    The cache maintains an additional connection per node (including replicas)
    in the cluster to listen to invalidation events
    """

    def __init__(
        self,
        connection_pool: BaseConnectionPool[ClusterConnection],
        cache: AbstractCache | None = None,
        max_idle_seconds: float = 15,
    ) -> None:
        """ """
        super().__init__(connection_pool, cache or LRUCache())
        self.node_caches: dict[TCPLocation, NodeTrackingCache[ClusterConnection]] = {}
        self._nodes: list[coredis.client.Redis[Any]] = []
        self._max_idle_seconds = max_idle_seconds

    def get_client_id(self, connection: ClusterConnection) -> int | None:
        if cache := self.node_caches.get(connection.location):
            return cache.client_id
        return None

    @property
    def healthy(self) -> bool:
        return all(cache.healthy for cache in self.node_caches.values())

    async def run(self, task_status: TaskStatus[None] = TASK_STATUS_IGNORED) -> None:
        self._nodes = [
            node.as_client(**self._connection_pool.connection_kwargs)
            for node in self._connection_pool.cluster_layout.nodes
        ]
        async with AsyncExitStack() as stack:
            nodes = []
            for node in self._nodes:
                nodes.append(await stack.enter_async_context(node))

            async with create_task_group() as tg:
                self._task_group = tg

                for node in nodes:
                    node_cache = NodeTrackingCache(
                        node.connection_pool,
                        cache=self._cache,
                        max_idle_seconds=self._max_idle_seconds,
                    )
                    await tg.start(node_cache.run)
                    assert node_cache._connection
                    self.node_caches[node_cache._connection.location] = node_cache
                task_status.started()



from __future__ import annotations

import warnings

from .patterns.cache import (
    AbstractCache,
    CacheStats,
    ClusterTrackingCache,
    LRUCache,
    NodeTrackingCache,
    TrackingCache,
)

warnings.warn(
    "coredis.patterns.cache has been moved to coredis.patterns.cache",
    DeprecationWarning,
    stacklevel=2,
)
__all__ = [
    "AbstractCache",
    "CacheStats",
    "LRUCache",
    "TrackingCache",
    "NodeTrackingCache",
    "ClusterTrackingCache",
]
